#include "stdafx.h"
#include "protocol.h"

using namespace protocol;

IPacket *ReadPacket(uint8 *data, size_t size, size_t &cursor)
{
	uint16 code = 0;
	uint8 type = 0;
	size_t cur = cursor;

	if(cur + 3 > size)
		return NULL;

	code = *(uint16 *)(data + cur);
	type = *(uint8 *)(data + cur + 2);
	cur += 3;

	IPacket *packet = NULL;

	if(type == 0)
	{
		packet = new CEmptyPacket;
		cursor = cur;
	}
	else if(type == 1)//byte
	{
		if(cur + 1 > size)
			return NULL;

		uint8 byte = *(uint8 *)(data + cur);
		cur += 1;

		packet = new CBytePacket(byte);
		cursor = cur;
	}
	else if(type == 2)//word
	{
		if(cur + 2 > size)
			return NULL;

		uint16 word = *(uint16 *)(data + cur);
		cur += 2;

		packet = new CWordPacket(word);
		cursor = cur;
	}
	else if(type == 3)//dword
	{
		if(cur + 4 > size)
			return NULL;

		uint32 dword = *(uint32 *)(data + cur);
		cur += 4;

		packet = new CDwordPacket(dword);
		cursor = cur;
	}
	else if(type == 4)//float
	{
		if(cur + 4 > size)
			return NULL;

		float f = *(float *)(data + cur);
		cur += 4;

		packet = new CFloatPacket(f);
		cursor = cur;
	}
	else if(type == 5)//string
	{
		if(cur + 2 > size)
			return NULL;

		uint16 len = *(uint16 *)(data + cur);
		cur += 2;

		if(cur + len > size)
			return NULL;

		CStringPacket *spacket = new CStringPacket;
		string &str = spacket->GetString();
		size_t index = 0;

		str.resize(len);

		for(string::iterator i = str.begin(); i != str.end(); i++)
			*i = *(data + cur + index++);

		cur += len;

		packet = spacket;
		cursor = cur;
	}
	else if(type == 6)//data
	{
		if(cur + 2 > size)
			return NULL;

		uint16 len = *(uint16 *)(data + cur);
		cur += 2;

		if(cur + len > size)
			return NULL;

		CDataPacket *dpacket = new CDataPacket;
		vector<uint8> &d = dpacket->GetData();
		size_t index = 0;

		d.resize(len);

		for(vector<uint8>::iterator i = d.begin(); i != d.end(); i++)
			*i = *(data + cur + index++);

		cur += len;

		packet = dpacket;
		cursor = cur;
	}
	else//unknown packet (probably corrupted data stream)
	{
		packet = new CUnknownPacket(type);
		cursor = cur;//shift 3 bytes header anycase
	}

	if(packet)
		packet->SetCode(code);

	return packet;
}

size_t CSocket::Recv(uint8 *data, size_t size, int &error)
{
	error = 0;

	int res = recv(mSocket, (char *)data, (int)size, 0);

	if(res == 0)//connection gracefully closed
	{
		error = 1;
		return 0;
	}
	else if(res == SOCKET_ERROR)
	{
		int err = WSAGetLastError();

		if(err == WSAEMSGSIZE)//buffer too small
		{
			return size;
		}
		else if(err == WSAEWOULDBLOCK)//no data available at the moment
		{
			return 0;
		}
		else//some other error
		{
			error = err;
			return 0;
		}
	}

	return res;
}

size_t CSocket::Send(uint8 *data, size_t size, int &error)
{
	error = 0;

	int res = send(mSocket, (const char *)data, (int)size, 0);

	if(res == SOCKET_ERROR)
	{
		int err = WSAGetLastError();

		if(err == WSAEWOULDBLOCK)//no buffer available on either client or host machine
		{
			return 0;
		}
		else
		{
			error = err;
			return 0;
		}
	}

	return res;
}

//Returns NULL on buffer underflow
IPacket *CSocket::ReadPacket(void)
{
	IPacket *packet;

	if(mInBuffer.size() != 0)
	{
		packet = ::ReadPacket(&(mInBuffer[0]), mInBuffer.size(), mInCursor);

		if(packet)
			return packet;
		else//   
		{
			if(mInCursor != 0)//     
				this->ShiftBuffer(mInBuffer, mInCursor);//      
		}
	}

	return NULL;
}

//Returns false on buffer overflow
bool CSocket::WritePacket(IPacket &packet)
{
	bool overflow = false;
	vector<uint8> data;
	packet.GetRawData(data);

	if(mOutBuffer.size() + data.size() > mOutBuffer.capacity())//overflow
	{
		//     
		if(mOutCursor != 0)
		{
			this->ShiftBuffer(mOutBuffer, mOutCursor);

			if(mOutBuffer.size() + data.size() > mOutBuffer.capacity())//   
				overflow = true;
		}
		else
			overflow = true;
	}
	else
	{
		//    

		size_t old_size = mOutBuffer.size();
		mOutBuffer.resize(mOutBuffer.size() + data.size());

		for(size_t i = 0; i < data.size(); i++)
			mOutBuffer[old_size + i] = data[i];
	}

	return !overflow;
}

bool CSocket::ProcessInput(int &error)
{
	error = 0;

	if(mInBuffer.capacity() - mInBuffer.size() == 0)
		return true;

	// --        --

	size_t old_size = mInBuffer.size();
	mInBuffer.resize(mInBuffer.capacity());
	int res = recv(mSocket, (char *)&(mInBuffer[old_size]), (int)(mInBuffer.capacity() - old_size), 0);

	if(res == 0)//connection gracefully closed
	{
		mInBuffer.resize(old_size);
		error = WSAGetLastError();

		if(error)
		{
			//error = error;
			return false;
		}
		else
		{
			error = 1;
			return false;
		}
	}
	else if(res == SOCKET_ERROR)
	{
		int err = WSAGetLastError();

		if(err == WSAEMSGSIZE)//buffer too small
		{
			//ok
			//the buffer remains resized to its capacity, as it has been filled till the cap
			return true;
		}
		else if(err == WSAEWOULDBLOCK)//no data available at the moment
		{
			mInBuffer.resize(old_size);
			return true;//buffer underflow
		}
		else//some other error
		{
			mInBuffer.resize(old_size);
			error = err;
			return false;
		}
	}

	mInBuffer.resize(old_size + res);

	return true;
}

bool CSocket::ProcessOutput(int &error)
{
	error = 0;

	// --        --

	if(mOutBuffer.size() - mOutCursor == 0)
		return true;
	
	int res = send(mSocket, (const char *)(&mOutBuffer[mOutCursor]), (int)(mOutBuffer.size() - mOutCursor), 0);

	if(res == SOCKET_ERROR)
	{
		int err = WSAGetLastError();

		if(err == WSAEWOULDBLOCK)//no buffer available on either client or host machine
		{
			return true;
		}
		else
		{
			error = err;
			return false;
		}
	}

	mOutCursor += res;

	return true;
}

void CSocket::ShiftBuffer(vector<uint8> &buffer, size_t &cursor)
{
	if(cursor == 0)
	{
		string msg = "The program was written by idiot\r\nReview line ";
		char buf[128] = "";
		_ltoa_s(__LINE__, buf, sizeof(buf), 10);
		msg += buf;
		msg += ", file ";
		msg += __FILE__;
		MessageBoxA(0, msg.c_str(), "", 0);
		return;
	}

	uint8 *ptr = &(buffer[0]);
	size_t num = buffer.size() - cursor;

	for(size_t i = 0; i < num; i++)
		*(ptr + i) = *(ptr + cursor + i);

	buffer.resize(buffer.size() - cursor);
	cursor = 0;
}

string protocol::GetHostByName(const string &name)
{
	struct hostent *hp;
	struct in_addr in;

	hp = gethostbyname(name.c_str());

	if(!hp || !*(hp->h_addr_list))
		return string("");
	
	memcpy(&in.s_addr, *(hp->h_addr_list), sizeof(in.s_addr));

	return string(inet_ntoa(in));
}

#include "protocol_error.inl"
