#pragma once

#include "../stdafx.h"

namespace protocol
{
	typedef signed char sint8;
	typedef unsigned char uint8;
	typedef signed short sint16;
	typedef unsigned short uint16;
	typedef signed long sint32;
	typedef unsigned long uint32;

	//: word   + byte   + 
	// : 0 - empty, 1 - byte, 2 - word, 3 - dword, 4 - float, 5 - string, 6 - data
	//: word   + length * byte' (   )

	class IPacket
	{
	public:

		IPacket(void)
		{
			mCode = 0;
		}

		virtual ~IPacket(void)
		{
		}

		void								SetCode(uint16 code)
		{
			mCode = code;
		}

		uint16								GetCode(void) const
		{
			return mCode;
		}

		uint8								GetType(void) const
		{
			return mType;
		}

		virtual void						GetRawData(vector<uint8> &data) = 0;

	protected:

		uint16								mCode;
		uint8								mType;
	};

	class CEmptyPacket : public IPacket
	{
	public:

		CEmptyPacket(void)
		{
			mType = 0;
		}

		virtual ~CEmptyPacket(void) {}

		virtual void						GetRawData(vector<uint8> &data)
		{
			data.resize(3);
			*(uint16 *)&(data[0]) = mCode;
			data[2] = 0;//type = empty
		}
	};

	class CBytePacket : public IPacket
	{
	public:

		CBytePacket(void)
		{
			mType = 1;
			mData = 0;
		}

		CBytePacket(uint8 data)
		{
			mType = 1;
			mData = data;
		}

		virtual ~CBytePacket(void) {}

		void								SetData(uint8 data)
		{
			mData = data;
		}

		uint8 &								GetData(void)
		{
			return mData;
		}

		const uint8 &						GetData(void) const
		{
			return mData;
		}

		virtual void						GetRawData(vector<uint8> &data)
		{
			data.resize(4);
			*(uint16 *)&(data[0]) = mCode;
			data[2] = 1;//type = byte
			data[3] = mData;
		}

	protected:

		uint8								mData;
	};

	class CWordPacket : public IPacket
	{
	public:

		CWordPacket(void)
		{
			mType = 2;
			mData = 0;
		}

		CWordPacket(uint16 data)
		{
			mType = 2;
			mData = data;
		}

		virtual ~CWordPacket(void) {}

		void								SetData(uint16 data)
		{
			mData = data;
		}

		uint16 &							GetData(void)
		{
			return mData;
		}

		const uint16 &						GetData(void) const
		{
			return mData;
		}

		virtual void						GetRawData(vector<uint8> &data)
		{
			data.resize(5);
			*(uint16 *)&(data[0]) = mCode;
			data[2] = 2;//type = word
			*(uint16 *)&(data[3]) = mData;
		}

	protected:

		uint16								mData;
	};

	class CDwordPacket : public IPacket
	{
	public:

		CDwordPacket(void)
		{
			mType = 3;
			mData = 0;
		}

		CDwordPacket(uint32 data)
		{
			mType = 3;
			mData = data;
		}

		virtual ~CDwordPacket(void) {}

		void								SetData(uint32 data)
		{
			mData = data;
		}

		uint32 &							GetData(void)
		{
			return mData;
		}

		const uint32 &						GetData(void) const
		{
			return mData;
		}

		virtual void						GetRawData(vector<uint8> &data)
		{
			data.resize(7);
			*(uint16 *)&(data[0]) = mCode;
			data[2] = 3;//type = dword
			*(uint32 *)&(data[3]) = mData;
		}

	protected:

		uint32								mData;
	};

	class CFloatPacket : public IPacket
	{
	public:

		CFloatPacket(void)
		{
			mType = 4;
			mData = 0;
		}

		CFloatPacket(float data)
		{
			mType = 4;
			mData = data;
		}

		virtual ~CFloatPacket(void) {}

		void								SetData(float data)
		{
			mData = data;
		}

		float &								GetData(void)
		{
			return mData;
		}

		const float &						GetData(void) const
		{
			return mData;
		}

		virtual void						GetRawData(vector<uint8> &data)
		{
			data.resize(7);
			*(uint16 *)&(data[0]) = mCode;
			data[2] = 4;//type = float
			*(float *)&(data[3]) = mData;
		}

	protected:

		float								mData;
	};

	//Can transfer up to 65535 symbols
	class CStringPacket : public IPacket
	{
	public:

		CStringPacket(void)
		{
			mType = 5;
		}
		
		CStringPacket(const string &str)
		{
			mType = 5;
			mString = str;
		}
		
		virtual ~CStringPacket(void) {}

		void								SetString(const string &str)
		{
			mString = str;
		}

		string &							GetString(void)
		{
			return mString;
		}

		const string &						GetString(void) const
		{
			return mString;
		}

		virtual void						GetRawData(vector<uint8> &data)
		{
			if(mString.length() > 0xffff)
				mString.resize(0xffff);

			data.resize(5 + mString.length());
			*(uint16 *)&(data[0]) = mCode;
			data[2] = 5;//type = string
			*(uint16 *)&(data[3]) = (uint16)mString.size();

			size_t index = 5;
			uint8 *ptr = &(data[0]);

			for(string::const_iterator i = mString.begin(); i != mString.end(); i++)
				*(ptr + index++) = (uint8)*i;
		}

	protected:

		string								mString;
	};

	//Can transfer up to 65535 bytes
	class CDataPacket : public IPacket
	{
	public:

		CDataPacket(void)
		{
			mType = 6;
		}
		
		CDataPacket(const vector<uint8> &data)
		{
			mType = 6;
			mData = data;
		}
		
		virtual ~CDataPacket(void) {}

		void								SetData(const vector<uint8> &data)
		{
			mData = data;
		}

		vector<uint8> &						GetData(void)
		{
			return mData;
		}

		const vector<uint8> &				GetData(void) const
		{
			return mData;
		}

		virtual void						GetRawData(vector<uint8> &data)
		{
			if(mData.size() > 0xffff)
				mData.resize(0xffff);

			data.resize(5 + mData.size());
			*(uint16 *)&(data[0]) = mCode;
			data[2] = 6;//type = data
			*(uint16 *)&(data[3]) = (uint16)mData.size();

			size_t index = 5;
			uint8 *ptr = &(data[0]);

			for(vector<uint8>::const_iterator i = mData.begin(); i != mData.end(); i++)
				*(ptr + index++) = *i;
		}

	protected:

		vector<uint8>						mData;
	};

	//Unknown packet indicates corrupted data stream
	class CUnknownPacket : public IPacket
	{
	public:

		CUnknownPacket(uint8 type)
		{
			mType = type;
		}

		virtual ~CUnknownPacket(void) {}

		virtual void						GetRawData(vector<uint8> &data)
		{
			data.resize(3);
			*(uint16 *)&(data[0]) = mCode;
			data[2] = mType;//type = unknown
		}
	};
	
	//      ,   NULL.
	IPacket *								ReadPacket(uint8 *data, size_t size, size_t &cursor);
	
	class CSocket
	{
	public:

		CSocket(void)
		{
			mSocket = INVALID_SOCKET;
			mInBuffer.reserve(102400);//TODO: move this to another rareer used place
			mOutBuffer.reserve(102400);
			mInCursor = mOutCursor = 0;
		}

		explicit CSocket(SOCKET s)
		{
			mSocket = s;
			mInBuffer.reserve(102400);
			mOutBuffer.reserve(102400);
			mInCursor = mOutCursor = 0;
		}

		CSocket(const CSocket &ref)
		{
			mSocket = ref.mSocket;
			mInBuffer.reserve(102400);
			mOutBuffer.reserve(102400);
			mInCursor = mOutCursor = 0;
		}

		const CSocket &						operator =(const CSocket &ref)
		{
			mSocket = ref.mSocket;
			mInBuffer.reserve(10240);
			mInBuffer.resize(0);
			mOutBuffer.reserve(10240);
			mOutBuffer.resize(0);
			mInCursor = mOutCursor = 0;

			return *this;
		}

		~CSocket(void)
		{
			//if(mSocket != INVALID_SOCKET)
			//	closesocket(mSocket);
			//Don't do such things ever!
		}

		bool								SetBlocking(bool blocking, int &error)
		{
			error = 0;

			unsigned long non_blocking = !blocking;
			
			if(ioctlsocket(mSocket, FIONBIO, &non_blocking) == SOCKET_ERROR)
			{
				error = WSAGetLastError();
				return false;
			}

			return true;
		}

		bool								Bind(const string &ip, uint16 port, int &error)
		{
			error = 0;

			if(mSocket != INVALID_SOCKET)
				this->Close();

			mSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

			if(mSocket == INVALID_SOCKET)
			{
				error = WSAGetLastError();
				return false;
			}

			sockaddr_in addr;
			addr.sin_family = AF_INET;
			addr.sin_addr.s_addr = inet_addr(ip.c_str());
			addr.sin_port = htons(port);

			if(bind(mSocket, (sockaddr *)&addr, sizeof(addr)) == SOCKET_ERROR)
			{
				closesocket(mSocket);
				mSocket = INVALID_SOCKET;
				return false;
			}

			return true;
		}
		
		bool								Listen(uint16 queue, int &error)
		{
			error = 0;

			if(listen(mSocket, queue) == SOCKET_ERROR)
			{
				error = WSAGetLastError();
				return false;
			}

			return true;
		}

		bool								Accept(CSocket &client_socket, int &error)
		{
			error = 0;

			SOCKET s;

			if((s = accept(mSocket, NULL, NULL)) == INVALID_SOCKET)
			{
				error = WSAGetLastError();
				return false;
			}

			client_socket = CSocket(s);

			return true;
		}

		bool								Connect(const string &ip, uint16 port, int &error)
		{
			error = 0;

			sockaddr_in addr;
			addr.sin_family = AF_INET;
			addr.sin_addr.s_addr = inet_addr(ip.c_str());
			addr.sin_port = htons(port);

			if(connect(mSocket, (sockaddr *)&addr, sizeof(addr)) == SOCKET_ERROR)
			{
				error = WSAGetLastError();
				return false;
			}

			return true;
		}

		void								Close()
		{
			if(mSocket != INVALID_SOCKET)
			{
				closesocket(mSocket);
				mSocket = INVALID_SOCKET;

				mInBuffer.resize(0);
				mOutBuffer.resize(0);
				mInCursor = mOutCursor = 0;
			}
		}

		//error == 1 - connection gracefully closed
		size_t								Recv(uint8 *data, size_t size, int &error);
		size_t								Send(uint8 *data, size_t size, int &error);
		
		//Returns NULL on buffer underflow
		//error == 0 - no error
		//error == 1 - connection gracefully closed
		IPacket *							ReadPacket(void);

		//Returns false on buffer overflow
		//error == 0 - no error
		//Take care of WSAENOBUFS! Dunno what's it.
		bool								WritePacket(IPacket &packet);

		//Tries to send as much data as possible
		bool								ProcessInput(int &error);
		bool								ProcessOutput(int &error);

	protected:

		//       
		void								ShiftBuffer(vector<uint8> &buffer, size_t &cursor);

		SOCKET								mSocket;
		
		vector<uint8>						mInBuffer;
		vector<uint8>						mOutBuffer;
		size_t								mInCursor;
		size_t								mOutCursor;
	};

	string GetHostByName(const string &name);

	char *GetErrorName(int error);
	char *GetErrorDesc(int error);
}
